/*-
 * Copyright (c) 2007
 * Tatsuya Kudoh(CDR/TK),ROYALPANDA.    All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package{
	import flash.display.*;
	import flash.events.*;
	import flash.geom.*;
	import flash.net.*;
	import flash.text.*;
	import flash.ui.*;
	import flash.utils.*;

	public class Game extends Sprite {
		public const left:int = 640;
		public const bottom:int = 480;
		public const saveObjName:String = "Royalpanda/fjump";

		protected var floorProb:Number;
		protected var floorNum:int;
		protected var floorX:int;
		protected var level:int;

		protected var bgOffset:Number = 0;
		protected var wallOffset:Number = 0;
		protected var scrollFraction:Number = 0;

		protected var floorOffset:Number;

		protected var scrollSpeed:Number;
		protected var scrollSpeedMin:Number;
		protected var scrollAccel:Boolean;

		protected var bg:Bitmap;
		protected var bgHeight:int;
		protected var leftWall:Bitmap;
		protected var rightWall:Bitmap;
		protected var player:Player;
		protected var score:int;
		protected var hiscore:int;

		protected var scoreLabel:Label;

		protected var joy:uint;
		protected var joyPrev:uint;

		protected var opFunction:Function;

		protected var elapsed:int;

		public function Game():void {
			var s:Sprite;
			var lab:Label;

			stage.scaleMode = StageScaleMode.NO_SCALE;

			s = new Sprite();
			s.graphics.beginFill(0x000040);
			s.graphics.drawRect(0,0,left,bottom);
			s.graphics.endFill();
			addChild(s);
			lab = new Label("Click to Start", 0xffffff, 0x000000, 24, "_");
			lab.x = 32;
			lab.y = 32;
			s.addChild(lab);

			s.addEventListener(MouseEvent.CLICK, function( e:Event ):void {
				removeChild(s);
				start();
			});
		}

		protected function start():void {
			var b:BitmapData;
			var bm:BitmapData;
			var r:Rectangle;
			var p:Point;

			hiscore = loadHiscore();

			b = Assets.bitmap(Assets.BM_BG);
			bm = new BitmapData(left, bottom + b.height, false);
			p = new Point(0,0);
			for( p.y = 0 ; p.y < bm.height ; p.y += b.height )
				for( p.x = 0 ; p.x < bm.width ; p.x += b.width )
					bm.copyPixels(b, b.rect, p);
			bg = new Bitmap(bm);
			bg.x = 0;
			bg.y = 0;
			addChild(bg);
			bgHeight = b.height;

			b = Assets.bitmap(Assets.BM_FLOOR);

			bm = new BitmapData(16, bottom+32);
			r = new Rectangle(0, 48, 16, 32);
			for( p = new Point(0,0) ; p.y < bm.height ; p.y += 32 )
				bm.copyPixels(b, r, p);
			leftWall = new Bitmap(bm);
			leftWall.x = 0;
			leftWall.y = 0;
			addChild(leftWall);

			bm = new BitmapData(16, bottom+32);
			r = new Rectangle(16, 48, 16, 32);
			for( p = new Point(0,0) ; p.y < bm.height ; p.y += 32 )
				bm.copyPixels(b, r, p);
			rightWall = new Bitmap(bm);
			rightWall.x = left - 16;
			rightWall.y = 0;
			addChild(rightWall);

			scoreLabel = new Label("0000", 0xffffc0, 0x000000, 20);
			scoreLabel.x = left - 32 - scoreLabel.width;
			scoreLabel.y = bottom - 24 - scoreLabel.height;
			addChild(scoreLabel);

			scrollRect = new Rectangle(0,0,left,bottom);

			joyPrev = 0;
			stage.addEventListener(KeyboardEvent.KEY_DOWN, keyListener);
			stage.addEventListener(KeyboardEvent.KEY_UP, keyListener);
			addEventListener(Event.ENTER_FRAME, enterFrameListener);

			elapsed = getTimer();
			doTitle();
		}

		protected function scroll( step:Number, add:Boolean = true ):void {
			var b:BitmapData;
			var i:int;
			var fl:Floor;
			var s:Number;

			bgOffset += step / 2;
			bgOffset %= bgHeight;
			bg.y = bgOffset - bgHeight;

			step +=scrollFraction;
			s = Math.floor(step);
			scrollFraction = step - s;

			wallOffset += s;
			wallOffset &= 31;
			leftWall.y = wallOffset - 32;
			rightWall.y = wallOffset - 32;

			if( player )
				player.y += s;

			for( i = 0 ; i < numChildren ; i++ ){
				fl = getChildAt(i) as Floor;
				if( fl ){
					fl.y += s;
					if( fl.y > bottom + 48 )
						removeChild(fl);
				}
			}

			if( add ){
				floorOffset += s;
				while( floorOffset >= 96 ){
					floorOffset -= 96;
					addFloor(floorOffset - 128);
				}
			}
		}

		protected function addFloor( aY:int ):void {
			var fl:Floor;
			var x0:int, x1:int;
			var w:int;
			var type:int;
			var prob:int;
			var f:Floor;
			var p:Floor;

			if( Rnd.coinToss() ){
				floorX += Rnd.next(65) + 128;
				if( floorX >= left - 64 )
					floorX = left - 128 - Rnd.next(65);
			}else{
				floorX -= Rnd.next(65) + 128;
				if( floorX < 64 )
					floorX = 128 + Rnd.next(65);
			}

			floorNum++;
			if( floorNum % 50 == 1 ){
				fl = new Floor(this, Floor.TYPE_NORMAL, left-32);
				fl.x = 16;
				floorProb += (100 - floorProb) * 0.25;

			}else{
				x0 = floorX - (Rnd.next(5)+5) * 8;
				x1 = floorX + (Rnd.next(5)+5) * 8;
				w = x1 - x0;
				if( x0 < 16 )
					x0 = 16;
				else if( x1 > left - 16 )
					x0 = left - 16 - w;

				prob = floorProb;

				if( Rnd.next(100) > prob - 30 ){
					fl = new Floor(this, Floor.TYPE_NORMAL, w);
				}else {
					if( Rnd.coinToss() ){
						fl = new Floor(this, Floor.TYPE_FROZEN, w);
					}else{
						fl = new Floor(this, Floor.TYPE_FORCE, w);
						fl.force = (Rnd.next(100 * level) / 200 + 2) * Rnd.sign();
					}
					prob *= 0.8;
				}

				if( Rnd.next(100) < prob - 30 ){
					fl.speed = (Rnd.next(100 * level) / 300 + 1) * Rnd.sign();
					prob *= 0.5;
				}

				if( Rnd.next(100) < prob - 50 ){
					fl.setBlinkDuration(Rnd.next(20) + 20);
				}
				fl.x = x0;
			}
			fl.floor = floorNum;
			fl.y = aY;
			addChildAt(fl, 1);
		}


		protected function showScore( s:int ):void {
			var i:int;
			var c:int;
			var str:String;

			if( s > 9999 )
				s = 9999;

			if( s > score ){
				score = s;
				str = "";
				for( i = 1000 ; i >= 1 ; i /= 10 ){
					c = s / i;
					s %= i;
					str += String.fromCharCode(c + 48);
				}
				scoreLabel.text = str;
				scoreLabel.draw();
			}
		}

		protected function loadHiscore():int {
			var s:SharedObject;

			s = SharedObject.getLocal(saveObjName);
			if( s == null )
				return 0;
			if( s.data.score is int )
				return s.data.score;

			return 0;
		}

		protected function saveHiscore( score:int ):void {
			var s:SharedObject;

			s = SharedObject.getLocal(saveObjName);
			if( s != null )
				s.data.score = score;
		}
			


		protected function doTitle():void {
			var i:int;
			var fl:Floor;
			var lab1:Label;
			var lab2:Label;
			var lab3:Label;
			var lab4:Label;
			var cnt:int;
			var start:Boolean;

			if( player ){
				removeChild(player);
				player = null;
			}
			for( i = numChildren-1 ; i >= 0 ; i-- ){
				fl = getChildAt(i) as Floor;
				if( fl )
					removeChild(fl);
			}

			lab1 = new Label("Falling Tower /",  0xffffff, 0x000020, 32, "_");
			addChild(lab1);
			lab1.x = (left - lab1.width) / 2;
			lab1.y = 128;

			lab2 = new Label("Version 0.0",  0xffffff, 0x000020, 16, "_");
			addChild(lab2);
			lab2.x = (left - lab2.width) / 2;
			lab2.y = 200;

			lab3 = new Label("Push Space Key",  0x00ffff, 0x000000, 16, "_");
			addChild(lab3);
			lab3.x = (left - lab2.width) / 2;
			lab3.y = 400;

			lab4 = new Label("HiScore " + hiscore,  0xffff40, 0x000000, 20);
			addChild(lab4);
			lab4.x = 32;
			lab4.y = bottom - 24 - lab4.height;

			score = 0;
			cnt = 0;
			start = false;

			opFunction = function( hit:uint ):void {
				scroll(1, false);
				cnt++;
				if( cnt == 12 ){
					cnt = 0;
					lab3.visible = ! lab3.visible;
				}
					
				if( hit & Joy.A ){
					start = true;
				}else if( start && ! (joy & Joy.A) ){
					removeChild(lab1);
					removeChild(lab2);
					removeChild(lab3);
					removeChild(lab4);
					doGame();
				}
			};
		}

		protected function doGame():void {
			var i:int;
			var gameOverCount:int;
			var lab1:Label;

			floorNum = 0;
			floorX = left / 2 + Rnd.next(128) * Rnd.sign();
			floorOffset = 0;

			floorProb = 0;
			level = 1;
			scrollSpeed = 800;
			scrollSpeedMin = 800 + scrollSpeed / 8;
			

			for( i = 352 ; i >= -128 ; i -= 96 )
				addFloor(i);

			player = new Player(this);
			player.x = (left - player.width)/2;
			player.y = 352 - player.height;
			addChild(player);

			scrollAccel = true;

			gameOverCount = 0;

			Snd.playBGM(Assets.SND_BGM);

			opFunction = function( hit:uint ):void {
				var i:int;
				var fl:Floor;
				var s:Number;
				var ss:Number;

				if( scrollAccel ){
					scrollSpeed *= 1.0015;
					if( scrollSpeed > 20000 ){
						scrollSpeed = 20000;
						scrollAccel = false;
					}
				}else{
					scrollSpeed *= 0.98;
					if( scrollSpeed <= scrollSpeedMin ){
						scrollSpeed = scrollSpeedMin;
						scrollSpeedMin = 800 + scrollSpeedMin / 8;
						if( scrollSpeedMin > 2000 )
							scrollSpeedMin = 2000;
						scrollAccel = true;
						if( level < 12 )
							level++;
					}
				}

				for( i = 0 ; i < numChildren ; i++ ){
					fl = getChildAt(i) as Floor;
					if( fl )
						fl.move();
				}
				if( gameOverCount == 0 ){
					player.move(joy);
					showScore(player.floor);
				}

				ss = scrollSpeed / 500;
				if( ss > 6 )
					ss = 6;
				s = 64 - player.y;
				if( s > 12 )
					s = 12;
				if( s < ss )
					s = ss;
				scroll(s);

				if( gameOverCount == 0 ){
					if( player.y > bottom + 48){
						gameOverCount++;
						Snd.stopBGM();
						Snd.play(Assets.SND_DEAD);
						if( score > hiscore ){
							hiscore = score;
							saveHiscore(score);
						}
					}
				}else{
					gameOverCount++;
					if( gameOverCount == 25 ){
						lab1 = new Label("Game Over",  0xffffff, 0x0000ff, 32, "_");
						addChild(lab1);
						lab1.x = (left - lab1.width) / 2;
						lab1.y = 128;
					}
					if( gameOverCount > 25 ){
						if( (hit & Joy.A) || gameOverCount > 200 ){
							removeChild(lab1);
							doTitle();
						}
					}
				}
			};
		}

		private function keyListener( e:KeyboardEvent ):void {
			var code:uint;

			code = Joy.fromKeyCode(e.keyCode);
			if( e.type == KeyboardEvent.KEY_DOWN )
				joy |= code;
			else
				joy &= ~code;
		}

		private function enterFrameListener( e:Event ):void {
			var hit:uint;

		/*
			var t:int;

			t = getTimer();
			while( elapsed < t ){
				elapsed += 40;
				hit = (joyPrev ^ joy) & joy;
				joyPrev = joy;
				opFunction(hit);
			}
		*/
			hit = (joyPrev ^ joy) & joy;
			joyPrev = joy;
			opFunction(hit);
		}
	}
}
